
// CodeGen.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2018

/*  CodeGen generates the Cabbage intermediate code,
    represented by the code and vars lists. Then it
    makes several passes over this data to generate 
    AE instructions, which it saves to a file.
*/


import java.io.*;
import java.util.*;


public class CodeGen
{
  // used to map longs into decimals
  private static final int RADIX_SZ = 8;        // 8 dp
  private static final int MAX_RADIX_SZ = 15;   // 15 dp

  private ArrayList<CodeLine> code;
  private ArrayList<VarInfo> vars;
  private int radixSz = RADIX_SZ;


  public CodeGen()
  {  code = new ArrayList<CodeLine>();
     vars = new ArrayList<VarInfo>();
  }


  public int getRadixSz()
  {  return radixSz;  } 


  public void setRadixSz(int r)
  {
    if ((r < 0) || ( r > MAX_RADIX_SZ))
      System.out.println("Radix size is out of range; using " + RADIX_SZ);
    else
      radixSz = r;
  }  // end of setRadixSz()



  // ------------ build intermediate code -----------------------


  public void addLine(CodeOps op, String varName)
  { 
    CodeLine cl = new CodeLine(op, varName);

    if ((op == CodeOps.LOAD) || (op == CodeOps.STORE) ||
        (op == CodeOps.STOREP) || (op == CodeOps.READ)) {
      VarInfo vi = new VarInfo(varName);
      addVarOnce(vi);
    }
    else if ((op == CodeOps.LEFT) || (op == CodeOps.RIGHT))
      cl.setValue( (long)parseDouble(varName));

    code.add(cl);
  }  // end of addLine(); 2 arg



  public void addNumber(CodeOps op, String varName, String valStr)
  {   addNumber(op, varName, valStr, true);  }   // number is shifted


  public void addNumber(CodeOps op, String varName, String valStr,
                                        boolean toShift)
  // for NUM and NUMA
  { 
    CodeLine cl = new CodeLine(op, varName);
    VarInfo vi = new VarInfo(varName);
    vi.setDefined();
    addVarOnce(vi);

    if (op == CodeOps.NUMA)
      cl.setArrayName(valStr);
    else if (op == CodeOps.NUM) {
      if (toShift)   // shift left
        cl.setValue( shiftLong(parseDouble(valStr)));
      else   // no shift
        cl.setValue( (long)parseDouble(valStr));
    }
    code.add(cl);
  }  // end of addNumber()


  public void addJumpLine(CodeOps op, int label)
  /* for adding lines tht include a jump label:
     JMP, JMPZ, and NOP (used as a destination)
  */
  { CodeLine cl = new CodeLine(op, null);
    cl.setLabel(label);
    code.add(cl);
  }


  public void addOp(CodeOps op)
  /* for adding op lines, such as ADD, JMPV, LOADPC,
     PUTCHAR, HALT, etc  */
  { CodeLine cl = new CodeLine(op, null);
    code.add(cl);
  }


  public void addArray(String nm, int size)
  // for adding a code line for a DIM instruction
  { 
    // System.out.println("ARR " + nm + " " + size);
    CodeLine cl = new CodeLine(CodeOps.NUM, nm);
    cl.setValue(0);
    code.add(cl);
    VarInfo vi = new VarInfo(nm, size);
    vi.setDefined();
    addVarOnce(vi);
  }


  private void addVarOnce(VarInfo vi)
  // only add this var info if it isn't already stored 
  { 
    if (findVar(vi.getID()) == null)
      vars.add(vi);
  }


  private VarInfo findVar(String id)
  // finf the var called id in the vars list
  {
    for (VarInfo vi : vars)
      if (vi.getID().equals(id))
        return vi;
    return null;
  }  


  private long shiftLong(double d)
  /* convert the double into a long by multiplying
     by 10^radix setting */
  { 
    double radix = Math.pow(10, radixSz);
    return (long)(Math.round(d*radix));
  }


  private double parseDouble(String s)
  {
    try {
      return (new Double(s)).doubleValue();
    }
    catch(NumberFormatException e) {
      System.out.println(s + " not a number; using 0");
    }
    return 0;
  }  // end of parseDouble()


  // ---------------- modify intermediate code ---------------


  public void derefForwards(HashMap<String, Integer> procLabels)
  /* Remove forward references by changing 
     JMPF <procID> lines into JMP label lines
  */
  {
    for(CodeLine cl : code) 
      if (cl.getOp() == CodeOps.JMPF) {
        cl.setOp(CodeOps.JMP);
        String procID = cl.getVar();
        Integer funcLabel = procLabels.get(procID);
        if (funcLabel == null) {
          System.out.println("No label found for procedure " + procID);
          cl.setLabel(-1);
        }
        else
          cl.setLabel((int)funcLabel);
      }
  }  // end of derefForwards()




  public void defineVars()
  /* add NUM definitions for the undefined vars. Place
     them at the start of the code, but after the external 
     function variables */
  { 
    for (VarInfo vi : vars) {
      if (!vi.isDefined() && isVarUsed(vi.getID())) {
        CodeLine cl = new CodeLine(CodeOps.NUM, vi.getID());
        cl.setValue(0);
        code.add(7, cl);
        vi.setDefined();
      }
    }
  }  // end of defineVars()



  private boolean isVarUsed(String id)
  {
    for(CodeLine cl : code) 
      if ( id.equals(cl.getVar()) )
        return true;
    return false;
  }  // end of isVarUsed()




  public void convertJMPs()
  /* Use intermediate code JMP instructions to generate
     AE combinatorial card instructions (e.g. CB, CF).
     This information is added to the code lines.
  */
  {
    int[] labelAddrs = new int[code.size()];

    // collect the code addresses of all the labels
    for(int i=0; i < code.size(); i++) {
      labelAddrs[i] = -1;   // set default value
      CodeLine cl = code.get(i);
      if ((cl.getOp() == CodeOps.NOP) &&
          (cl.getLabel() != -1))
         labelAddrs[i] = cl.getLabel();   // addr i has this label
    }

    // convert the jumps into combinatorial cards
    for(int i=0; i < code.size(); i++) {
      CodeLine cl = code.get(i);
      if ((cl.getOp() == CodeOps.JMPZ) ||
          (cl.getOp() == CodeOps.JMP)) {
        String cmd = null;
        int jumpAddress = getJumpAddress(labelAddrs, cl.getLabel());
        if (jumpAddress != -1) {
          int move = jumpAddress - (i+1);
          if (move < 0) {
            cmd = "CB";
            move = -move;
          }
          else
            cmd = "CF";
          if (cl.getOp() == CodeOps.JMPZ)
            cmd = cmd + "?";
          else
            cmd = cmd + "+";
          cmd = cmd + move;
        }
        else
          cmd = "  NOP since error in JMP label";
        cl.setCard(cmd);
      }
    }
  }  // end of convertJMPs()


  private int getJumpAddress(int[] labelAddrs, int label)
  {
    for(int i=0; i < labelAddrs.length; i++)
      if (labelAddrs[i] == label)
        return i;
    System.out.println("Label " + label + " not found");
    return -1;
  }  // end of getJumpAddress()


  public void setVarAddrs()
  /* Pass over the code and calculate the addresses of the
     variables in the store. Take into account extra space
     needed for arrays.
     Store that info as part of their VarInfo data.
  */
  { int addr = 1;   // first address in the store
    for(CodeLine cl : code) {
      if ((cl.getOp() == CodeOps.NUM) ||
          (cl.getOp() == CodeOps.NUMA)) {
        String id = cl.getVar();
        VarInfo vi = findVar(id);
        if (vi == null)
          System.out.println("Could not find variable " + id);
        else {
          vi.setAddress(addr);
          int size = vi.getSize();
          if (size > 0)
            addr += size;
          else
            addr++;
        }
      }
    }
  }  // end of setVarAddrs()



  public void genCards()
  /* Convert all the intermediate code ops into
     AE instructions. Store that information in the
     code lines using setCard().
  */
  {
    for(CodeLine cl : code) {
      VarInfo vi = findVar(cl.getVar());  // may be null
      CodeOps op = cl.getOp();
      if (op == CodeOps.NOP)
        cl.setCard("     NOP");
      else if (op == CodeOps.NUM) {
        cl.setCard("N" + vi.getAddress() + " " + cl.getValue());
                  // N<address> <value>
      }
      else if (op == CodeOps.NUMA) {
        VarInfo viArray = findVar(cl.getArrayName());
        long shiftedArrAddr = shiftLong( viArray.getAddress());
           // shifted so addition of shifted offset is correct
        cl.setCard("N" + vi.getAddress() + " " + shiftedArrAddr);
      }
      else if (op == CodeOps.LOAD)
        cl.setCard("L" + vi.getAddress());
      else if (op == CodeOps.LOADI)      // new card
        cl.setCard("l" + vi.getAddress());
      else if (op == CodeOps.LOADPC)     // new card
        cl.setCard("v");
      else if (op == CodeOps.STORE)
        cl.setCard("S" + vi.getAddress());
      else if (op == CodeOps.STOREI)     // new card
        cl.setCard("s" + vi.getAddress());
      else if (op == CodeOps.STOREP)
        cl.setCard("S" + vi.getAddress() + "\'");
      else if (op == CodeOps.STOREIP)    // new card
        cl.setCard("s" + vi.getAddress() + "\'"); 
      else if (op == CodeOps.ADD)
        cl.setCard("+");
      else if (op == CodeOps.SUB)
        cl.setCard("-");
      else if (op == CodeOps.MULT)
        cl.setCard("*");
      else if (op == CodeOps.DIV)
        cl.setCard("/");
      else if ((op == CodeOps.JMPZ) || 
               (op == CodeOps.JMP)) {}   
           // done earlier in convertJMPs()
      else if (op == CodeOps.JMPV)       // new card
        cl.setCard("J");
      else if (op == CodeOps.PRINT)
        // cl.setCard("P");
        cl.setCard("D " + radixSz);      // new card
      else if (op == CodeOps.PUTCHAR)    // new card
        cl.setCard("U");
      else if (op == CodeOps.READ)       // new card
        cl.setCard("R" + vi.getAddress() + " " + radixSz);
      else if (op == CodeOps.EXT)    // new card
        cl.setCard("X " + radixSz);
      else if (op == CodeOps.RIGHT)
        cl.setCard(">" + cl.getValue());
      else if (op == CodeOps.LEFT)
        cl.setCard("<" + cl.getValue());
      else if (op == CodeOps.HALT)
        cl.setCard("H");
      else {
        System.out.println("Unrecognized op: " + op);
        cl.setCard("     ?? for " + op);
      }
    }
  }  // end of genCards()


  public void saveCards(String fnm)
  /* Write out the AE instructions to a file. Include
     other code line info as comments.
  */
  { System.out.println();
	  try (PrintWriter out = 
                  new PrintWriter(new FileWriter(fnm))) {
      for(CodeLine cl : code)
        out.println(cl.getCard() + "   \t\t\t." + cl.getDetails(false));
      System.out.println("Saved cards to " + fnm);
		} 
    catch (IOException e) 
    {  System.out.println("Could not save cards to " + fnm);  }
  }  // end of saveCards()



  public void showCode()
  { 
    System.out.println("\nCode: ");
    for(CodeLine cl : code)
      System.out.println(cl);
    System.out.println();
  }  // end of showCode()


  public void showVars()
  {
    System.out.println("\nVars:");
    for (VarInfo vi : vars)
      System.out.println("  " + vi);
    System.out.println();
  }  // end of showVars()

}  // end of CodeGen class
