
// Parser.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Jan 2018

/* Recursive descent parser which generates intermediate
   code which is converted to AE instructions by the
   CodeGen class.
*/

import java.io.*;
import java.util.*;


public class Parser
{
  private BufferedReader br;
  private Lexer lexer;   // lexical analyzer
  private Token tok;        // current token

  private CodeGen cg;

  private int jumpNo = 0;   // for creating jump labels
  private int varCount = 0; // for creating variable names

  private String currFuncID = "main";  
      /* name of function currently being processed;
         the top-level code is called main by default */

  private HashMap<String, Integer> funcLabels = new HashMap<>(); 
      // functions in the code and their jump labels

  private ArrayList<String> paramNames = new ArrayList<>();
      /* stores the parameter names of the function currently
         being processed, so they can be renamed differently */

  private ArrayList<String> constNames = new ArrayList<>();
      /* names of any constants used in the program;
         this list prevents them from being renamed */



  public Parser(String fnm, CodeGen cg) throws IOException
  { 
    this.cg = cg;
    br = new BufferedReader(new FileReader(fnm));
    lexer = new Lexer(br);
    getToken();
  }


  public HashMap<String, Integer> getFuncLabels()
  {  return funcLabels;  }


  private void getToken() throws IOException
  {  tok = lexer.nextToken(); }



  public void program() throws IOException
  // EBNF: <program> ::= [ <numdps> ] <main> { <func> }
  {
    if (tok.symbol() == Toks.NUMDPS)
      numdps();

  /*
     The 7 variables used for external function calls
     are "func_Nm", "func_nargs" for the number of arguments, and
     "func_arg0" to "func_arg3"  will hold at most 4 arguments.
     The result of the call will be stored in "func_result".
     They must stay at the start of the cards.
  */
    cg.addNumber(CodeOps.NUM, "func_Nm", "0");
    cg.addNumber(CodeOps.NUM, "func_nargs", "0");
    cg.addNumber(CodeOps.NUM, "func_arg0", "0");
    cg.addNumber(CodeOps.NUM, "func_arg1", "0");
    cg.addNumber(CodeOps.NUM, "func_arg2", "0");
    cg.addNumber(CodeOps.NUM, "func_arg3", "0");
    cg.addNumber(CodeOps.NUM, "func_result", "0");

    // useful constants
    cg.addNumber(CodeOps.NUM, "zero", "0");
    cg.addNumber(CodeOps.NUM, "one", "1");
    cg.addNumber(CodeOps.NUM, "e", "2.71828183");
    cg.addNumber(CodeOps.NUM, "pi", "3.14159265");

    constNames.add("zero"); constNames.add("one");
    constNames.add("e"); constNames.add("pi");

    main();
    while (tok.symbol() != Toks.EOF)  
      func(); 
    br.close();
  }  // end of program()


  public void numdps() throws IOException
  // EBNF: <numdps> ::= "numdps" "=" <number> ";"
  {
    match(tok, Toks.NUMDPS); 
    match(tok, Toks.ASSIGN);
    if (tok.symbol() == Toks.NUMBER) {
      int num_dps = (new Integer(tok.lexeme())).intValue();
      cg.setRadixSz(num_dps);
      getToken(); 
    }
    else 
      error(lexer.position(), "Expected a number");
    match(tok, Toks.SEMICOLON); 
  }  // end of numdps()



  public void main() throws IOException
  // EBNF: <main> ::= <stat> { <stat> }
  {
    currFuncID = "main";    // this is the 'main' function
    stat();
    while ((tok.symbol() != Toks.EOF) &&
           (tok.symbol() != Toks.FUNC))
      stat(); 
    cg.addOp(CodeOps.HALT);    // just to be safe
  }  // end of main()




  public void func() throws IOException
  /* EBNF: <func> ::= 
              "func" <ident> "(" [ <ident> { "," <ident> } ] ")"
              "{" <stat> { <stat> } "}"
  */
  {
    match(tok, Toks.FUNC); 
    currFuncID = id();
    int funcLabel = jumpNo++;

    cg.addJumpLine(CodeOps.NOP, funcLabel);   
             // label the function start so jmp can aim for somewhere
    cg.addLine(CodeOps.NOP, "func " + currFuncID);   
              // included as documentation in the AE code
    if (funcLabels.get(currFuncID) != null)
      System.out.println("Function " + currFuncID + "already defined");
    else
      funcLabels.put(currFuncID, funcLabel);

    paramNames.clear();
    match(tok, Toks.LPAREN);
    if (tok.symbol() != Toks.RPAREN) {
      paramNames.add( id());
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        paramNames.add( id());
      }
    }
    match(tok, Toks.RPAREN);

    match(tok, Toks.BEGIN);
    stat();
    while (tok.symbol() != Toks.END)
      stat(); 
    match(tok, Toks.END);

    doReturn();   
  }  // end of func()


  private void doReturn()
  /* at runtime do a jump using an offset calculated from
     the return address and the current PC address
  */
  { cg.addOp(CodeOps.SUB);  
    cg.addLine(CodeOps.LOAD, "ret_" + currFuncID);
    cg.addOp(CodeOps.LOADPC);
    cg.addOp(CodeOps.JMPV);
  }  // end of doReturn()



  private void stat() throws IOException
  /* EBNF: <stat> ::= 
                <assign> ";"
              | if "(" <cond> ")" <stat> [ else <stat> ]
              | "{" <stat> { <stat> } "}"
              | while "(" <cond> ")" <stat>
              | read <var> ";"
              | print ( <expr> | <string> ) 
                        { "," ( <expr> | <string> ) } ";"
              | dim <ident> "[ <number> "]" ";"
              | <call> ";"
              | return <expr> ";"
              | e         // empty statement
  */
  { 
    String id, indexVar;
    int endLabel;

    switch (tok.symbol()) {

    case ID:
    case LPAREN:
      assign();
      match(tok, Toks.SEMICOLON); 
      break;

    case IF:
      getToken();
      match(tok, Toks.LPAREN);
      cond();
      match(tok, Toks.RPAREN);
      int elseLabel = jumpNo++;
      cg.addJumpLine(CodeOps.JMPZ, elseLabel);   // to else if 0
      stat();
      endLabel = jumpNo++;
      cg.addJumpLine(CodeOps.JMP, endLabel);    // to end
      cg.addJumpLine(CodeOps.NOP, elseLabel);   
                      // include even if no else branch
      if (tok.symbol() == Toks.ELSE) { 
        getToken(); 
        stat();
      } 
      cg.addJumpLine(CodeOps.NOP, endLabel);
      break;

    case BEGIN:
      getToken();
      stat();
      while (tok.symbol() != Toks.END)
        stat(); 
      match(tok, Toks.END); 
      break;

    case WHILE:
      getToken();
      int whileLabel = jumpNo++;
      cg.addJumpLine(CodeOps.NOP, whileLabel);
      match(tok, Toks.LPAREN);
      cond();
      match(tok, Toks.RPAREN);
      endLabel = jumpNo++;
      cg.addJumpLine(CodeOps.JMPZ, endLabel);   // to end if 0
      stat();
      cg.addJumpLine(CodeOps.JMP, whileLabel);
      cg.addJumpLine(CodeOps.NOP, endLabel);
      break;

    case READ:
      getToken();
      id = tok.lexeme();
      indexVar = varIndex();
      if (indexVar != null) {
        String tempVar = "read" + varCount++;
        cg.addLine(CodeOps.READ, tempVar);
        cg.addOp(CodeOps.ADD);
        cg.addLine(CodeOps.LOAD, tempVar);
        cg.addLine(CodeOps.LOAD, "zero");
        storeIndex(id, indexVar, false);
      }
      else   // numerical var
        cg.addLine(CodeOps.READ, renameVar(id));
      match(tok, Toks.SEMICOLON); 
      break;

    case PRINT:   
      // print (<expr>|<string>) {"," (<expr>|<string>) } ";"
      getToken();
      printItem();
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        printItem();
      }
      // add code for printing a newline char
      cg.addNumber(CodeOps.NUM, "ch", "10", false); 
              // do not shift numbers representing chars
      cg.addOp(CodeOps.ADD);
      cg.addLine(CodeOps.LOAD, "ch"); 
      cg.addLine(CodeOps.LOAD, "zero");
      cg.addOp(CodeOps.PUTCHAR);

      match(tok, Toks.SEMICOLON); 
      break;

    case DIM:
      getToken();
      id = id();
      match(tok, Toks.LBRACE); 
      if (tok.symbol() == Toks.NUMBER) {
        int size = (new Integer(tok.lexeme())).intValue();
        cg.addArray(id, size);
        cg.addNumber(CodeOps.NUMA, id+"_Address", id);
           /* this instruction will be changed into a NUM
              to hold the start address of the array when it
              is known */
        getToken();
        match(tok, Toks.RBRACE); 
        match(tok, Toks.SEMICOLON); 
      }
      else
        error(lexer.position(), "Number expected");
      break;

    case CALL:   // call Cabbage function, or
    case EXT:    // call an external function
      ecall();
      match(tok, Toks.SEMICOLON); 
      break;


    case RETURN:
      getToken();
      String exprID = expr();
      match(tok, Toks.SEMICOLON); 
      copyTo("result_" + currFuncID, exprID);
      doReturn();
      break;

    default:   // to allow a statment to be empty
      
    }  // end of switch
  }  // end of stat()



  private String renameVar(String var)
  /* Renames a variable to include its function name.
     This enforces a simple form of lexical scoping so
     a variable cannot be used outside the function.
  */
  {
    int pos = constNames.indexOf(var);
    if (pos != -1)  // it is a constant name
      return var;

    pos = paramNames.indexOf(var);
    if (pos == -1)   // not a function parameter
      return currFuncID + "_" + var;
    else
      return currFuncID + "_" + pos;
  }  // end of renameVar()



  private void printItem()  throws IOException  
  // deals with the BNF for (<expr>|<string>) in the print statement
  {
    if (tok.symbol() == Toks.STRING) {
      String msg = tok.lexeme().replaceAll("^\"|\"$", "");
              // remove first and last double quotes
      for(int i=0; i < msg.length(); i++) {
        cg.addNumber(CodeOps.NUM, "ch", 
               "" + ((int)msg.charAt(i)), false);   
               // no shift for chars
        cg.addOp(CodeOps.ADD);
        cg.addLine(CodeOps.LOAD, "ch");
        cg.addLine(CodeOps.LOAD, "zero");
        cg.addOp(CodeOps.PUTCHAR);
      }
      getToken(); 
    }
    else {  // it is an expression
      String exprID = expr();  
      cg.addOp(CodeOps.ADD);  // force mill processing
      cg.addLine(CodeOps.LOAD, exprID);
      cg.addLine(CodeOps.LOAD, "zero");
      cg.addOp(CodeOps.PRINT);
    }
  }  // end of printItem()



  private String ecall() throws IOException
  // call a Cabbage or an external function
  /* EBNF: <call> ::= ( "call" | "ext" ) 
               <ident> "(" [ <expr> { "," <expr> } ] ")"
  */
  { if (tok.symbol() == Toks.CALL)
      return cabbageCall();
    else if (tok.symbol() == Toks.EXT)
      return externalCall();
    else {
      error(lexer.position(), "call or ext expected");
      return null;
    }
  }  // end of ecall()



  private String cabbageCall() throws IOException
  // EBNF: "call" <ident> "(" [ <expr> { "," <expr> } ] ")"
  {
    int numArgs = 0;
    match(tok, Toks.CALL); 
    String funcID = id();
    match(tok, Toks.LPAREN);
    if (tok.symbol() != Toks.RPAREN) {
      String exprID = expr();
      copyTo(funcID + "_" + numArgs++, exprID);
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        exprID = expr();
        copyTo(funcID + "_" + numArgs++, exprID);
      }
    }
    match(tok, Toks.RPAREN);
    addCallJumps(funcID);
    return "result_"+funcID;
  }  // end of cabbageCall()



  private void copyTo(String storeNm, String loadNm)
  // storeNm gets a copy of loadNm
  {
    cg.addOp(CodeOps.ADD); 
    cg.addLine(CodeOps.LOAD, loadNm);
    cg.addLine(CodeOps.LOAD, "zero");
    cg.addLine(CodeOps.STORE, storeNm);
  }  // end of copyTo()



  private void copyStrTo(String storeNm, String valStr)
  // storeNm gets a copy of the valStr string
  {
    String tempVar = "tempVar"+varCount++;
    cg.addNumber(CodeOps.NUM, tempVar, valStr, false); 
        // value string numbers are not shifted
    copyTo(storeNm, tempVar);
  }  // end of copyStrTo()





  private void addCallJumps(String funcID)
  /* add code to jump to a function and return to
     the next line when the function returns */
  {
    Integer funcLabel = funcLabels.get(funcID);

    cg.addOp(CodeOps.ADD); // store pc in function's return jump var
    cg.addOp(CodeOps.LOADPC); 
    cg.addLine(CodeOps.LOAD, "zero");
    cg.addLine(CodeOps.STORE, "ret_" + funcID);
 
    // add code to jump to the start of the function
    if (funcLabel == null)  
      /* this means the call is a forward reference; the
         function has not been defined yet in the program */
      cg.addLine(CodeOps.JMPF, funcID);  
      /* the JMPF will be converted to a JMP later after
         the function has been defined */
    else
      cg.addJumpLine(CodeOps.JMP, (int)funcLabel);

    cg.addLine(CodeOps.NOP, "ret_" + funcID);  
      /* the return jumps to this NOP as a safe place to 
         resume after the end of the call */
  }  // end of addCallJumps()



  private String externalCall() throws IOException
  /* EBNF: "ext" <ident> "(" [ <expr> { "," <expr> } ] ")"

     The 7 variables used for external function calls
     are "func_Nm", "func_nargs" for the number of arguments, and
     "func_arg0" to "func_arg3"  will hold at most 4 arguments.
     The result of the call will be stored in "func_result".
     They must stay at the start of the cards.
  */
  {
    int numArgs = 0;
    match(tok, Toks.EXT); 
    String funcID = id();
    String fid = encodeName(funcID);
    copyStrTo("func_Nm", fid);   // func_Nm is assigned fid

    match(tok, Toks.LPAREN);
    if (tok.symbol() != Toks.RPAREN) {
      String exprID = expr();
      copyTo("func_arg" + numArgs++, exprID);
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        exprID = expr();
        copyTo("func_arg" + numArgs++, exprID);
      }
    }

    copyStrTo("func_nargs", ""+numArgs);

    cg.addOp(CodeOps.EXT);
    match(tok, Toks.RPAREN);
    return "func_result";
  }  // end of externalCall()



  private String encodeName(String funcID)
  // convert the name into a string of ASCII digits
  {
    StringBuilder sb = new StringBuilder();
    for (int i=0; i < funcID.length(); i++)
      sb.append("" + encodeChar(funcID.charAt(i)));
    // System.out.println("Encoded ID: " + sb);
    return sb.toString();
  } // end of encodeName()


  private int encodeChar(char ch)
  /* Convert the ch into an ASCII value.
     Only lowercase letters, digits, and '_' are
     allowed. Lowercase letters are converted to
     uppercase so that all the ASCII values will be
     2-digit numbers.
  */
  { if ((ch >= 'a') && (ch <= 'z'))
      return (int)(Character.toUpperCase(ch));

    if (((ch >= '0') && (ch <= '9')) || (ch == '_'))
      return (int)ch;

    System.out.println("Illegal letter; using \'x\'");
    return (int)'X';
  }  // end of encodeChar()



  private void assign() throws IOException
  /* EBNF:  <assign> ::=  
                      <var> "=" <call> | <expr>
                    | <assignDiv>
  */
  { if (tok.symbol() == Toks.ID) {  // start of <var>
      String id = tok.lexeme();
      String indexVar = varIndex();
      match(tok, Toks.ASSIGN);

      if ((tok.symbol() == Toks.CALL) ||
          (tok.symbol() == Toks.EXT)) {  // start of <call>
        String resultID = ecall();
        cg.addOp(CodeOps.ADD); 
        cg.addLine(CodeOps.LOAD, resultID);
        cg.addLine(CodeOps.LOAD, "zero");
      }
      else {
        String exprID = expr();
        cg.addOp(CodeOps.ADD);  // load expr result into mill
        cg.addLine(CodeOps.LOAD, exprID);
        cg.addLine(CodeOps.LOAD, "zero");
      }

      if (indexVar != null)
        storeIndex(id, indexVar, false);
      else   // ordinary var is renamed
        cg.addLine(CodeOps.STORE, renameVar(id));
    }
    else if (tok.symbol() == Toks.LPAREN)  // start of division
      assignDiv();
    else 
      error(lexer.position(), "Unrecognizable symbol");
  }  // end of assign()



  private void assignDiv() throws IOException
  /* EBNF:  <assignDiv> ::=
                  "(" <quotient var> "," <remainder var> ")"  
                   "=" <argum> "/" <argum>
  */
  { match(tok, Toks.LPAREN);
    String idQ = tok.lexeme();  // var for quotient
    String indexQ = varIndex();

    match(tok, Toks.COMMA);
    String idR = tok.lexeme();  // var for remainder
    String indexR = varIndex();

    match(tok, Toks.RPAREN);
    match(tok, Toks.ASSIGN);

    cg.addOp(CodeOps.DIV);
    String leftID = argum();
    cg.addLine(CodeOps.LOAD, leftID);
    cg.addLine(CodeOps.LEFT, ""+cg.getRadixSz());
    match(tok, Toks.DIV);
    String rightID = argum(); 
    cg.addLine(CodeOps.LOAD, rightID);

    if (indexQ != null)
      storeIndex(idQ, indexQ, true);
    else   // ordinary var is renamed
      cg.addLine(CodeOps.STOREP, renameVar(idQ));

    if (indexR != null)
      storeIndex(idR, indexR, false);
    else   // ordinary var is renamed
      cg.addLine(CodeOps.STORE, renameVar(idR));
  }  // end of assignDiv()



  private void storeIndex(String id, String indexVar, 
                                           boolean isPrime)
  /* This implements A[i] = expr, where id contains
     the 'A' name, indexVar is the i value, and expr
     is in one of the output registers of the mill (as
     indicated by the isPrime boolean).
  */
  { cg.addLine(CodeOps.STORE, "exprVal");  
           // temporarily store the expression result

    // calculate address of A + i, and store in tempVar
    cg.addOp(CodeOps.ADD);
    cg.addLine(CodeOps.LOAD, id+"_Address");
    cg.addLine(CodeOps.LOAD, indexVar);
    cg.addLine(CodeOps.RIGHT, ""+cg.getRadixSz());
           // undo shift to get a small index integer
    String tempVar = "Array_" + varCount++;
    cg.addLine(CodeOps.STORE, tempVar);

    // reload the expression value
    cg.addOp(CodeOps.ADD);
    cg.addLine(CodeOps.LOAD, "exprVal");
    cg.addLine(CodeOps.LOAD, "zero");

    // store the expr value in the address loc 
    // stored in tempVar
    if (isPrime)
      cg.addLine(CodeOps.STOREIP, tempVar);
    else
      cg.addLine(CodeOps.STOREI, tempVar);
  }  // end of storeIndex()



  private String argum() throws IOException
  // EBNF: <argum> ::= <var> | <number>
  {
    if (tok.symbol() == Toks.ID) {
      String id = tok.lexeme();
      String indexVar = varIndex(); 
      if (indexVar != null)
        return loadIndex(id, indexVar);
      else    // ID is a variable name
        return renameVar(id);
    }
    else if (tok.symbol() == Toks.NUMBER) {
      String numStr = tok.lexeme();
      getToken(); 
      String tempVar = "argum" + varCount++;
      cg.addNumber(CodeOps.NUM, tempVar, numStr);
      return tempVar;
    }
    else {
      error(lexer.position(), "Unrecognizable symbol");
      return null;
    }
  }  // end of argum()


  private String loadIndex(String id, String indexVar)
  /* This loads A[i] into a temporary variable.
     id contains the 'A' name, and indexVar is 
     the i value.
  */
  { // calculate address of A + i, and store in tempVar
    cg.addOp(CodeOps.ADD);
    cg.addLine(CodeOps.LOAD, id+"_Address");
    cg.addLine(CodeOps.LOAD, indexVar);
    cg.addLine(CodeOps.RIGHT, ""+cg.getRadixSz());
    String tempVar = "Array_" + varCount++;
    cg.addLine(CodeOps.STORE, tempVar);

    cg.addOp(CodeOps.ADD);  // indirect load of A+i
    cg.addLine(CodeOps.LOADI, tempVar);
    cg.addLine(CodeOps.LOAD, "zero");
    cg.addLine(CodeOps.STORE, tempVar);  // reuse tempVar
    return tempVar;
  }  // end of loadIndex()



  private String varIndex() throws IOException
  /* A variable may be a simple name or an array
     reference. The EBNF: 
        <var> ::= <ident> [ "[" ( <ident> | <number> ) "]" ]

     The name for <ident> has already been obtained before
     this call, so this code only returns a variable containing
     the array index, or null if this variable is simple.
  */
  { String indexVar = null;
    id();   // id name already obtained earlier
    if (tok.symbol() == Toks.LBRACE) {
      match(tok, Toks.LBRACE); 
      if (tok.symbol() == Toks.ID)
        indexVar = renameVar(id());
      else if (tok.symbol() == Toks.NUMBER) {
        String numStr = tok.lexeme();
        getToken(); 
        indexVar = "var" + varCount++;
        cg.addNumber(CodeOps.NUM, indexVar, numStr);
         // store the number in a new variable; return that var
      }
      else
        error(lexer.position(), "Unrecognizable index");
      match(tok, Toks.RBRACE); 
    }
    return indexVar;
  }  // end of varIndex()


  private String id() throws IOException
  {
    String id = null;
    if (tok.symbol() == Toks.ID) {
      id = tok.lexeme();
      getToken();
    }
    else
      error(lexer.position(), Toks.ID.getText() + " expected");
    return id;
  }  // end of id()



  private String expr() throws IOException
  /* EBNF: <expr> ::= 
            [ ("+"|"-") ] <term> { ("+"|"-") <term> }
  */
  {
    String leftTermVar = null;
    if ((tok.symbol() == Toks.PLUS) || 
        (tok.symbol() == Toks.MINUS)) { 
      Toks op = tok.symbol();
      getToken(); 
      String termVar = term();

      if (op == Toks.PLUS)
        cg.addOp(CodeOps.ADD);
      else
        cg.addOp(CodeOps.SUB);
      cg.addLine(CodeOps.LOAD, "zero");
      cg.addLine(CodeOps.LOAD, termVar);

      leftTermVar = "expr" + varCount++;
      cg.addLine(CodeOps.STORE, leftTermVar);
    } 
    else
      leftTermVar = term();

    String rightTermVar = null;
    while ((tok.symbol() == Toks.PLUS) || 
           (tok.symbol() == Toks.MINUS)) {
      Toks op = tok.symbol();
      getToken(); 
      rightTermVar = term();
      if (op == Toks.PLUS)
        cg.addOp(CodeOps.ADD);
      else
        cg.addOp(CodeOps.SUB);
      cg.addLine(CodeOps.LOAD, leftTermVar);
      cg.addLine(CodeOps.LOAD, rightTermVar);
      leftTermVar = "expr" + varCount++;
      cg.addLine(CodeOps.STORE, leftTermVar);
    }

    // cg.addLine(CodeOps.LOAD, leftTermVar);
    return leftTermVar;
  }  // end of expr()



  private String term() throws IOException
  // EBNF: <term> ::= <factor> { ( "*" | "/" ) <factor> }
  {
    String rightFVar = null;

    String leftFVar = factor();
    while ((tok.symbol() == Toks.TIMES) ||
           (tok.symbol() == Toks.DIV)) { 
      Toks op = tok.symbol();
      getToken(); 
      rightFVar = factor(); 

      if (op == Toks.TIMES) {
        cg.addOp(CodeOps.MULT);
        cg.addLine(CodeOps.LOAD, leftFVar);
        cg.addLine(CodeOps.LOAD, rightFVar);
        leftFVar = "term" + varCount++;
        cg.addLine(CodeOps.RIGHT, ""+cg.getRadixSz());
        cg.addLine(CodeOps.STORE, leftFVar);
      }
      else {  // a div
        cg.addOp(CodeOps.DIV);
        cg.addLine(CodeOps.LOAD, leftFVar);
        cg.addLine(CodeOps.LEFT, ""+cg.getRadixSz());
        cg.addLine(CodeOps.LOAD, rightFVar);
        leftFVar = "term" + varCount++;
        cg.addLine(CodeOps.STOREP, leftFVar);
             // ignore the remainder part inside an expr
     }
    }
    return leftFVar;
  }  // end of term()




  private String factor() throws IOException
  // EBNF:  <factor> ::=  <argum> |  "(" <expr> ")"
  {
    if ((tok.symbol() == Toks.ID)|| 
        (tok.symbol() == Toks.NUMBER))
      return argum();
    else if (tok.symbol() == Toks.LPAREN) {
      getToken(); 
      String id = expr();
      match(tok, Toks.RPAREN); 
      return id;
    }
    else {
      error(lexer.position(), "Unrecognizable symbol");
      return null;
    }
  }  // end of factor()



  private void cond() throws IOException
  // EBNF: <cond> ::= <expr1> ">=" <expr2> 
  {
    String expr1Var = expr();
    match(tok, Toks.GE);
    String expr2Var = expr();

    /* e1 >= e2 is implemented as e1 - e2 which
       may trigger a run up if e1 is positive and
       the (e1-e2) result is negative
    */
    cg.addOp(CodeOps.SUB);
    cg.addLine(CodeOps.LOAD, expr1Var);
    cg.addLine(CodeOps.LOAD, expr2Var); 
  }  // end of cond()



  private void match(Token t, Toks match) throws IOException
  {
    if (tok.symbol() == match) 
      getToken(); 
    else 
      error(lexer.position(), match.getText() + " expected");
  }  // end of match()



  private void error(int pos, String msg)
  {
    System.out.println();
    for (int i = 0; i < pos; i++) 
      System.out.print(" ");
    System.out.println("^");
    System.out.println("***** Error: " + msg + " *****");
    System.exit(1);
  }  // end of error()


}  // end of Parser class
