
// ParserTest.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2018

/* A recursive descent parser which doesn't generate
   anything (i.e. no parse tree or code), but is useful
   for testing the implementation of the Cabbage 
   grammar rules.
*/

import java.io.*;


public class ParserTest
{
  private BufferedReader br;
  private Lexer lexer;   // lexical analyzer
  private Token tok;        // current token


  public ParserTest(String fnm) throws IOException
  { 
    br = new BufferedReader(new FileReader(fnm));
    lexer = new Lexer(br);
    getToken();
  }


  private void getToken() throws IOException
  {  tok = lexer.nextToken(); }



  public void program() throws IOException
  // EBNF: <program> ::= [ <numdps> ] <main> { <func> }
  {
    if (tok.symbol() == Toks.NUMDPS)
      numdps();
    main();
    while (tok.symbol() != Toks.EOF)  
      func(); 
    br.close();
  }  // end of program()


  public void numdps() throws IOException
  // EBNF: <numdps> ::= "numdps" "=" <number> ";"
  {
    match(tok, Toks.NUMDPS); 
    match(tok, Toks.ASSIGN);
    if (tok.symbol() == Toks.NUMBER) {
      int num_dps = (new Integer(tok.lexeme())).intValue();
      // System.out.println(num_dps);
      getToken(); 
    }
    else 
      error(lexer.position(), "Expected a number");
    match(tok, Toks.SEMICOLON); 
  }  // end of numdps()



  public void main() throws IOException
  // EBNF: <main> ::= <stat> { <stat> }
  {
    stat();
    while ((tok.symbol() != Toks.EOF) &&
           (tok.symbol() != Toks.FUNC))
      stat(); 
  }  // end of main()


  public void func() throws IOException
  // EBNF: <func> ::= func <ident> "(" [ <ident> { "," <ident> } ] ")"
  //                                 "{" <stat> { <stat> } "}"
  {
    match(tok, Toks.FUNC); 
    match(tok, Toks.ID);

    match(tok, Toks.LPAREN);
    if (tok.symbol() != Toks.RPAREN) {
      match(tok, Toks.ID);
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        match(tok, Toks.ID);
      }
    }
    match(tok, Toks.RPAREN);

    match(tok, Toks.BEGIN);
    stat();
    while (tok.symbol() != Toks.END)
      stat(); 
    match(tok, Toks.END);
  }  // end of func()


  private void stat() throws IOException
  /* EBNF: <stat> ::= 
                <assign> ";"
              | if "(" <cond> ")" <stat> [ else <stat> ]
              | "{" <stat> { <stat> } "}"
              | while "(" <cond> ")" <stat>
              | read <var> ";"
              | print ( <expr> | <string> ) 
                   { "," ( <expr> | <string> ) } ";"
              | dim <ident> "[ <number> "]" ";"
              | <call> ";"
              | return <expr> ";"
              | e
  */
  { 
    switch (tok.symbol()) {
    case ID:
    case LPAREN:
      assign();
      match(tok, Toks.SEMICOLON); 
      break;

    case IF:
      getToken();
      match(tok, Toks.LPAREN);
      cond();
      match(tok, Toks.RPAREN);
      stat();
      if (tok.symbol() == Toks.ELSE) { 
        getToken(); 
        stat();
      } 
      break;

    case BEGIN:
      getToken();
      stat();
      while (tok.symbol() != Toks.END) {
        stat(); 
        // System.out.println("E");
      }
      match(tok, Toks.END); 
      break;

    case WHILE:
      getToken();
      match(tok, Toks.LPAREN);
      cond();
      match(tok, Toks.RPAREN);
      stat();
      break;

    case READ:
      getToken();
      var();
      match(tok, Toks.SEMICOLON); 
      break;

    case PRINT:    // print ( <expr> | <string> ) { "," ( <expr> | <string> ) }
      getToken();
      if (tok.symbol() == Toks.STRING)
        getToken(); 
      else
        expr();

      while (tok.symbol() == Toks.COMMA) {
        getToken();
        if (tok.symbol() == Toks.STRING)
          getToken(); 
        else
          expr();
      }
      match(tok, Toks.SEMICOLON); 
      break;

    case DIM:
      getToken();
      match(tok, Toks.ID);
      match(tok, Toks.LBRACE); 
      match(tok, Toks.NUMBER);
      match(tok, Toks.RBRACE); 
      match(tok, Toks.SEMICOLON); 
      break;

    case CALL:
    case EXT:
      call();
      match(tok, Toks.SEMICOLON); 
      break;

    case RETURN:
      getToken();
      expr();
      match(tok, Toks.SEMICOLON); 
      break;

    default:   // for e
      
    }
  }  // end of stat()


  private void assign() throws IOException
  /* EBNF: 
     <assign> ::=  <var> "=" ( <expr>  | <call> )
         |     "(" <var> "," <var> ")"  "=" <argum> "/" <argum>
  */
  { if (tok.symbol() == Toks.ID) {
      var();
      match(tok, Toks.ASSIGN);
      if (tok.symbol() == Toks.CALL)
        call();
      else
        expr();
    }
    else if (tok.symbol() == Toks.LPAREN){
      getToken();
      var();
      match(tok, Toks.COMMA);
      var();
      match(tok, Toks.RPAREN);
      match(tok, Toks.ASSIGN);
      argum();
      match(tok, Toks.DIV);
      argum(); 
    }
    else 
      error(lexer.position(), "Unrecognizable symbol");
  }  // end of assign()



  private void call() throws IOException
  /* EBNF: <call> ::= ( "call" | "ext" )
        <ident> "(" [ <expr> { "," <expr> } ] ")"
  */
  {
    if ((tok.symbol() == Toks.CALL) ||
        (tok.symbol() == Toks.EXT)) 
      getToken(); 
    else
      error(lexer.position(), "call or ext expected");
    match(tok, Toks.ID);
    match(tok, Toks.LPAREN);
    if (tok.symbol() != Toks.RPAREN) {
      expr();
      while (tok.symbol() == Toks.COMMA) {
        getToken();
        expr();
      }
    }
    match(tok, Toks.RPAREN);
  }  // end of call()


  private void argum() throws IOException
  // EBNF: <argum> ::= <var> | <number>
  {
    if (tok.symbol() == Toks.ID) 
      var(); 
    else if (tok.symbol() == Toks.NUMBER)
      getToken(); 
    else 
      error(lexer.position(), "Unrecognizable symbol");
  }  // end of argum()



  private void var() throws IOException
  /* EBNF: <var> ::= <ident> 
            [ "[" ( <ident> | <number> ) "]" ]
  */
  {
    if (tok.symbol() == Toks.ID) {
      getToken();
      // System.out.println("ID");
      if (tok.symbol() == Toks.LBRACE) {
        match(tok, Toks.LBRACE); 
        if ((tok.symbol() == Toks.ID) ||
            (tok.symbol() == Toks.NUMBER))
          getToken(); 
        else
          error(lexer.position(), "Unrecognizable index");
        match(tok, Toks.RBRACE); 
      }
    }
    else 
      error(lexer.position(), "Unrecognizable symbol");
  }  // end of var()



  private void expr() throws IOException
  /* EBNF: <expr> ::= [ <op> ] <term> { <op> <term> }
           <op> ::= "+" | "-"
  */
  {
    if ((tok.symbol() == Toks.PLUS) || 
        (tok.symbol() == Toks.MINUS)) { 
      getToken(); 
      // System.out.println("PLUS");
    } 
    term();
    while ((tok.symbol() == Toks.PLUS) || 
           (tok.symbol() == Toks.MINUS)) {
      getToken(); 
      term(); 
    }
  }  // end of expr()



  private void term() throws IOException
  // EBNF: <term> ::= <factor> { ("*" | "/") <factor> }
  {
    factor();
    while ((tok.symbol() == Toks.TIMES) ||
           (tok.symbol() == Toks.DIV)) { 
      getToken(); 
      factor(); 
    }
  }  // end of term()



  private void factor() throws IOException
  /* EBNF:  <factor> ::=  <argum> |  "(" <expr> ")"
  */
  {  // System.out.println("FCT");
    if ((tok.symbol() == Toks.ID)|| 
        (tok.symbol() == Toks.NUMBER))
      argum();
    else if (tok.symbol() == Toks.LPAREN) {
      getToken(); 
      expr();
      match(tok, Toks.RPAREN); 
    }
    else 
      error(lexer.position(), "Unrecognizable symbol");
  }  // end of factor()



  private void cond() throws IOException
  // EBNF: <cond> ::= <expr1> ">=" <expr2> 
  {
    expr();
    match(tok, Toks.GE);
    expr();
  }  // end of cond()




  private void match(Token t, Toks match) throws IOException
  {
    if (tok.symbol() == match) 
      getToken(); 
    else 
      error(lexer.position(), match.getText() + " expected");
  }  // end of match()



  private void error(int pos, String msg)
  {
    System.out.println();
    for (int i = 0; i < pos; i++) 
      System.out.print(" ");
    System.out.println("^");
    System.out.println("***** Error: " + msg + " *****");
    System.exit(0);
  }  // end of error()


  // ----------------------------------------------

  public static void main(String args[]) throws java.io.IOException
  {
    if (args.length != 1) {
      System.out.println("Usage: java ParserTest <CB fnm>");
      return;
    }
    ParserTest parser = new ParserTest(args[0]);
    parser.program();
  }  // end of main()


}  // end of ParserTest class
