
// AES.java

/* Fourmilab's Analytical Engine Simulator, by John Walker
   https://www.fourmilab.ch/babbage/contents.html

   Cut down functionality by:
   Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2018

  - Removed support for:
      * ALL Attendant Request Cards:
         Calculation Trace Cards
         Numeric Output Format Cards
         Output Annotation Cards
         Advancing and Backing Block Cards
         Card Library Inclusion Requests
         Decimal Place Expansion Cards

      * Curve Drawing Cards
      * The GUI and Web/applet support

      For details on these, see
      https://www.fourmilab.ch/babbage/cards.html

  - Replaced BigInt by Java's BigInteger
  - used ArrayList<> instead of Vector


   Loads cards from a file named on the command line,
   starts the Engine and displays the results on standard
   output.
  
   New cards:
  
    // nicer I/O (not essential)
    *   (R | r ) <var column> [<shift left dist>]     // read
    *   U          // putchar
    *   D <shift dp>       // print as a double with dp shifted left
  
    // to support arrays
    *   l, z, s    // indirect versions of L, Z, S
                   
    // for subroutine support
    *   v     // load the current pc value into the mill
    *   j     // jump offset loaded from the mill
  
    // for calling external Java methods
    * X  <shift dp>
  
   ----------------------
   For a good description of card types see,
  
    https://cs.stanford.edu/people/eroberts/courses/soco/
                            projects/1998-99/babbage/index.htm

   Top-level:
    https://cs.stanford.edu/people/eroberts/courses/soco/
                            projects/babbage/FormattedWell.html

  ------
  Ada 2012 emulation of Charles Babbages Analytical Engine:
    https://github.com/simonjwright/analytical-engine
  and
    https://cp4space.wordpress.com/2016/02/06/
              deep-learning-with-the-analytical-engine/

  Usage:

   > javac *.java
   > java AES [-t] div.ae
  or as a JAR:
   > jar -cvfm aes.jar manifest.txt *.class
   > java -jar aes.jar -t div.ae
*/


import java.util.*;
import java.math.*;
import java.lang.reflect.*;


public class AES
{
  private final static BigInteger K10e50 = new BigInteger(
      "100000000000000000000000000000000000000000000000000");
      // a memory cell can hold up to 50 decimal digits, 
      // so this is a max value for a number

  private final static int STEP_MAX = 100;
    // max number of shifts that can be applied to a number

  private final static int RACK_SIZE = 1000;
    // the number of memory cells stored in the mill's rack

  private final static int READ_SHIFT_MAX = 20;
    // added by ad
    /* max number of shifts that can be applied to a number
       read from input. This is small so the shifted number
       can be encoded as a Java long */



  private CardReader reader;
  private Store store;   // similar to RAM
  private Mill mill;     // similar to an ALU

  private boolean isTracing = false;
  private boolean isRunning = true;


  public AES(String fnm, boolean t)
  {
    isTracing = t;
    reader = new CardReader(fnm, isTracing);
    mill = new Mill(isTracing);
    store = new Store(isTracing);

    System.out.println("Executing...");
    while (isRunning)
      processCard();
  }  // end of AES()



  private void processCard()
  {
    String cmd = reader.getCard();
    if (cmd == null) {
      isRunning = false;
      return;
    }

    if (isTracing)
      System.out.println("Card:  " + cmd);

    char op = (cmd.length() == 0) ? ' ' : cmd.charAt(0);
    boolean isPrime = false;
    boolean inChain = true;
    int n = 0;
    int nIndirect = 0;
    BigInteger v = BigInteger.ZERO;
    String[] toks;

    // strip possible comment from end of cmd
    int loc = 0;
    if ((loc = cmd.indexOf(". ", 1)) >= 0)
      /* It's okay to use trim() here since the only
         situation in which leading space would be
         dropped is if the card is a comment in the
         first place.  Since the op has already
         been extracted, no harm is done if the leading
         space on the comment is dropped by the trim()
         below. */
      cmd = cmd.substring(0, loc).trim();

    switch (op) {

    // Mill operations (Operation cards; arithmetic instructions)
    case '+':
    case '-':
    case '*':   // ad: only allow ASCII chars for op names
    case '/':
      mill.setOp(op);
      break;

    case '<':    // shift a number left or right
    case '>':
      n = -1;
      if (cmd.length() > 1) {
        try {
          n = Integer.parseInt(cmd.substring(1));
        }
        catch (NumberFormatException ne) {}
      }
      if ((n < 0) || (n > STEP_MAX))
        errorHalt("Bad stepping up/down card", cmd);
      mill.shiftAxes((op == '<') ? n : -n);
      break;


    // Mill to store transfers (Variable cards; load/store ops)
    case 'L':
    case 'Z':
    case 'S':
    case 'l':    // indirect versions added by ad
    case 'z':
    case 's':
      isPrime = false;
      try {
        String addrStr = cmd.substring(1).trim();
        if (addrStr.endsWith("'")) {
          isPrime = true;
          addrStr = addrStr.substring(0, addrStr.length()-1).trim();
        }
        n = Integer.parseInt(addrStr);
      }
      catch (NumberFormatException ne)
      {  n = -1; }

      if (n < 0) {
        errorHalt("Bad variable card", cmd);
        break;
      }

      switch (op) {
      case 'L':   // store --> mill
        mill.transferIn(store.get(n), isPrime);
        break;

      case 'Z':
        mill.transferIn(store.get(n), isPrime);
        store.set(n, BigInteger.ZERO);
        break;

      case 'S':  // mill --> store
        store.set(n, mill.transferOut(isPrime));
        break;

      // ad's new cards: indirect versions of L, Z, and S
      case 'l':
        nIndirect = (store.get(n)).intValue();
        mill.transferIn(store.get(nIndirect), isPrime);
        break;

      case 'z':
        nIndirect = (store.get(n)).intValue();
        mill.transferIn(store.get(nIndirect), isPrime);
        store.set(nIndirect, BigInteger.ZERO);
        break;

      case 's':
        nIndirect = (store.get(n)).intValue();
        store.set(nIndirect, mill.transferOut(isPrime));
        break;
      }

      break;


    // Number card; similar to load immediate
    case 'N':
      toks = cmd.substring(1).split("\\s+");
      n = -1;
      if (toks.length >= 2) {
        try {
          n = Integer.parseInt(toks[0]);   // address
          if (toks[1].charAt(0) == '+')
            toks[1] = toks[1].substring(1);
          v = new BigInteger(toks[1]);     // value
        }
        catch (NumberFormatException ne) {}
      }
      if ((n < 0) || (n > RACK_SIZE-1) ||
          (v.abs().compareTo(K10e50) >= 0) ) {
        errorHalt("Bad number card", cmd);
        break;
      }
      store.set(n, v);
      break;


    // Combinatorial cards; control transfer ops;
    // Format:  C (F | B) (+ | ? ) <number>
    case 'C': {
        int dist;   // no. of cards to jump over
        inChain = true;
        if ((cmd.length() < 4) ||
            ((cmd.charAt(1) != 'F') && (cmd.charAt(1) != 'B')) ||
            ((cmd.charAt(2) != '?') && (cmd.charAt(2) != '+')) ) {
          errorHalt("Bad combinatorial card", cmd);
          break;
        }
        try {
          dist = Integer.parseInt(cmd.substring(3).trim());
        }
        catch (NumberFormatException ne) {
          dist = -1;
        }
        if (n < 0) {
          errorHalt("Bad combinatorial card value", cmd);
          break;
        }

        if ((cmd.charAt(2) == '+') || mill.hasRunUp()) {
          if (cmd.charAt(1) == 'F')
            inChain = reader.jumpForward(dist);
          else
            inChain = reader.jumpBack(dist);
          if (!inChain) {
            errorHalt("Card chain exhausted", cmd);
            break;
          }
        }
      }
      break;


    // control cards
    case 'B': // ring bell
      System.out.print("\007");
      System.out.flush();
      break;

    case 'P': // print with a newline
      mill.println();
      break;
/*
    case 'p': // print without a newline (new by ad)
      mill.print();
      break;
*/
    case 'H': // halt
      isRunning = false;
      break;

    case ' ':
    case '.': // Comment
      break;


    // -------------- new cards by ad ---------------------


    // read card    (R | r) <address> [<shift left>]
    case 'R':
    case 'r':    // read indirect
      toks = cmd.substring(1).split("\\s+");
      n = -1;
      int shift = 0;   // default is no shift
      if (toks.length >= 1) {
        try {
          n = Integer.parseInt(toks[0]);        // address
          if (toks.length >= 2)
            shift = Integer.parseInt(toks[1]);
        }
        catch (NumberFormatException ne) 
        {  System.out.println(ne); }
      }
      if ((n < 0) || (n > RACK_SIZE-1) ||
          (shift < 0) || (shift > READ_SHIFT_MAX) ) {
        errorHalt("Bad read card", cmd);
        break;
      }
      // System.out.println("read " + n + " " + shift);
      BigInteger readVal = read(toks[0], shift);
      // System.out.println("  " + readVal);

      if (op == 'R')   // input --> store
        store.set(n, readVal);
      else {  // 'r'
        nIndirect = (store.get(n)).intValue();
        store.set(nIndirect, readVal);
      }
      break;


    case 'U': // pUtchar
      mill.putchar();
      break;


    case 'D': // print as a double with dp shifted left
      try {
        n = Integer.parseInt( cmd.substring(1).trim());
      }
      catch (NumberFormatException ne)
      {  n = 0; }
      mill.printDouble(n);
      break;


    // l, z, s : indirect versions of L, Z, and S (see above)


    case 'J':    //  the jump offset is loaded from the mill
      int offset = (mill.transferOut(false)).intValue() + 3;
                // add to adjust loadpc locations
      // System.out.println("offset: " + offset);
      if (offset < 0)
        inChain = reader.jumpBack(-offset);
      else
        inChain = reader.jumpForward(offset);
      if (!inChain)
        errorHalt("Card chain exhausted", cmd);
      break;


    case 'v':   // for loading pc into the mill
      int pc = reader.getIndex();
      mill.transferIn(BigInteger.valueOf(pc), false);
      break;


    case 'X':     
    /*  call external Java method;
        treat arguments and result as doubles with 
        their dp shifted left by n
    */
      try {
        n = Integer.parseInt( cmd.substring(1).trim());
      }
      catch (NumberFormatException ne)
      {  n = 0; }
      extractFuncCall(n);
      break;


    // -------------------------------------

    default:
      errorHalt("Unknown op", cmd);
      break;
    }
  }  // end of processCard()




  private void errorHalt(String msg, String cmd)
  {
    System.out.println(msg + ": \"" + cmd + "\"");
    isRunning = false;
  }


  private BigInteger read(String varID, int shift)
  /* The shifted number is represented by a long, so the
     shift is restricted to READ_SHIFT_MAX
  */
  {
    BigInteger val = BigInteger.ZERO;
    try {
      Scanner scan = new Scanner(System.in);
      System.out.print("  ? ");
      double d = Double.parseDouble(scan.nextLine());
      long l = (long)(Math.round(d* Math.pow(10,shift)));
      val = BigInteger.valueOf(l);
    } 
    catch (Exception e) 
    { System.out.println(e);
      System.out.println("Using 0");
    }
    return val;
  }  // end of read()


  // ----------------- execute external methods -------------


  private void extractFuncCall(int numDPs)
  /* The 7 variables used for external function calls
     are "func_Nm", "func_nargs" for the number of arguments, and
     "func_arg0" to "func_arg3"  will hold at most 4 arguments.
     The result of the call will be stored in "func_result".
     They are located at the start of the cards, 1-7.  */
  {
/*  for (int i=1; i < 8; i++)
      System.out.println("Card " + i + ":" + store.get(i));
*/
    String fnVal = (store.get(1)).toString();
    if (fnVal.equals("0")) {
      System.out.println("No function found");
      return;
    }

    String funcNm = decodeName(fnVal);
    // System.out.println("funcName: " + fnVal + "; " + funcNm);

    int numArgs = (store.get(2)).intValue();
    if (numArgs > 4) {
      System.out.println("No. of arguments too large; reading 4");
      numArgs = 4;
    }
    // System.out.println("numArgs: " + numArgs);
   
    Object[] vals = new Double[numArgs];
    for (int i=0; i < numArgs; i++) {
      BigInteger bi = store.get(i+3);  // cards 3-6
      vals[i] = (new BigDecimal(bi, numDPs)).doubleValue();
      // System.out.println("arg" + i + ": " + vals[i]);
    }

    double result = callFunc(funcNm, vals);
    // System.out.println("result: " + result);

    long l = (long)(Math.round(result* Math.pow(10,numDPs)));
    BigInteger val = BigInteger.valueOf(l);
    // System.out.println("BigInt: " + val);

    store.set(7, val);  // store on results card
    // System.out.println("Card 7: " + store.get(7));
  }  // end of extractFuncCall()



  private String decodeName(String nameDigits)
  /* The loop assumes that each ASCII decimal digit is 
     encoded by two characters, which means that
     only the ASCII for uppercase letters (65-90)
     can be used.
     Most lowercase letters use 3 digits (i.e. 'd' and later).
  */
  {
    char[] nameChars = new char[nameDigits.length()/2];
    for(int i=0; i< nameChars.length; i++) {
      String chDigits = nameDigits.substring(i*2, i*2+2);
      nameChars[i] = decodeChar(chDigits);
    }
    return new String(nameChars);
  } // end of decodeName()


  private char decodeChar(String chDigits)
  /* only consider uppercase letters, digits, and _
     but convert uppercase to lowercase since 
     function names look 'nicer' written this way
  */
  {
    char ch = 'X'; 
    try {
      ch = (char) Integer.parseInt(chDigits);
    }
    catch (NumberFormatException ne) {}

    if ((ch >= 'A') && (ch <= 'Z'))
      return Character.toLowerCase(ch);

    if (((ch >= '0') && (ch <= '9')) || (ch == '_'))
      return ch;

    System.out.println("Illegal letter; using \'x\'");
    return 'x';
  }  // end of decodeChar()



  private double callFunc(String funcNm, Object[] vals)
  // execute external function (if possible)
  {
	  try {
      //load the Utils class
	  	Class<?> utilsCls = Class.forName("Utils");

      // access the method
	    Class[] params = new Class[vals.length];
      for (int i=0; i < vals.length; i++)
        params[i] = Double.TYPE;
      Method method = utilsCls.getDeclaredMethod(funcNm, params);

	  	// call the method
	  	double res = ((Double) method.invoke(null, vals)).doubleValue();
      return res;
	  }
    catch(Exception e)
    {  System.out.println("Could not execute " + funcNm + ":\n" + e); 
       return 0;
    }
   }  // end of callFunc()

  // --------------------------------------------


  public static void main(String[] args)
  {
   if (args.length == 0) {
     System.out.println("Usage: java [-t] <file>.ae");
     return;
   }

   boolean isTracing = false;
   String fnm = null;
   if (args.length == 1)
     fnm = args[0];
   else {
     isTracing = (args[0].equals("-t"));
     fnm = args[1];
   }

   new AES(fnm, isTracing);
  }  // end of main()



}  // end of AES class
