
// Mill.java

/* Fourmilab's Analytical Engine, by John Walker
   https://www.fourmilab.ch/babbage/contents.html

   Cut down functionality by:
   Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2018

   The mill performs the same role as a modern ALU.

   For a good description of the mill see,
   https://cs.stanford.edu/people/eroberts/courses/soco/projects/1998-99/babbage/index.htm
*/

import java.util.*;
import java.math.*;


public class Mill
{
  private final static BigInteger K10e50 = new BigInteger(
      "100000000000000000000000000000000000000000000000000");
      // a memory cell can hold up to 50 decimal digits
  private final static BigInteger Km10e50 = K10e50.negate();

  private static final int MAX_SHIFT = 101;  
     // a number can be shifted 0 cells, up to a maximum of 100


  private enum OPS {   // mill operations
      NONE, ADD, SUBTRACT, MULTIPLY, DIVIDE;
  }

  private BigInteger[] ingress = new BigInteger[3];  // 3 inputs
          // two axes and a primed axis
          // an ingress 'axis' corresponds to an input register

  /* The Primed Axis is used to compute the quotient of a 100 digit 
     number divided by a 50 digit number. The result 
     is a 50 digit number. 

     The first 50 digits of the 100 digit number is  
     stored on Ingress Axis 1, and the second 50 digits 
     on the Primed Axis.
  */


  private BigInteger[] egress = new BigInteger[2];   // 2 outputs
          // one axis and a primed axis
          // an egress 'axis' corresponds to an output register

  /* The Primed Axis is used during division. 
     The quotient is placed on the Primed Axis and 
     the remainder on the Egress Axis. 
     For example, 20 / 7, would put 2 on the Primed Axis, 
     and 6 on the Egress Axis. 

     If the quotient is larger than 50 digits or the divisor is zero, 
     the run-up lever is set. This functionality is at the heart of 
     conditional jumps
  */


  private BigInteger axisVal = ingress[0];
          // may hold an ingress or egress axis value

  private int axisIdx = 0;  // the index of the current ingress axis


  private BigInteger[] shiftFactor;  // for the < and > ops
    /* holds powers of ten in BigInteger form, so they
       don't need to be calculated more than once */


  private OPS op = OPS.NONE;   // for the mill operations

  private boolean isTracing = false;

  private boolean runUp = false;    // the run-up lever
  /* The run up lever is used to implement tests in Cabbage's 
     if's and while's. 
     The lever switches to its up position (true) in two cases :

     1. The sign on the result of a sum is different from the 
        sign on the first number in the sum, on Ingress Axis 1. 
        For example, consider the sum (4  9). This will cause 
        the run up lever to be set because the result (-5) has 
        a different sign then Ingress Axis 1 (the number 4.) 
        This feature is used by Cabbage's if's and while's

     2. The sum's result is greater than 50 digits long. 
        For example, if you divide a 100 digit number by 
        a 2 digit number, the run up lever is set. 
        This overflow detection is not used by Cabbage.
   */


  public Mill(boolean t)
  {
    shiftFactor = new BigInteger[MAX_SHIFT];
    for(int i=0; i < MAX_SHIFT; i++)
      shiftFactor[i] = BigInteger.ONE;

    op = OPS.NONE;
    for (int i = 0; i < 3; i++)
      setIngress(i, BigInteger.ZERO);

    for (int i = 0; i < 2; i++)
      setEgress(i, BigInteger.ZERO);

    axisVal = BigInteger.ZERO;  // dummy value
    axisIdx = 0;
    runUp = false;
    isTracing = t;
  }  // end of Mill()



  private void setIngress(int idx, BigInteger v)
  // store v in input register idx
  {  ingress[idx] = v;  }

  private void setEgress(int idx, BigInteger v)
  // store v in output register idx
  {  egress[idx] = v;  }


  public boolean hasRunUp()
  // is the run up lever set?
  {  return runUp;  }


  public void setOp(char opChar) 
                      throws IllegalArgumentException
  {
    switch (opChar) {   // only allow ASCII chars for op names
    case '+':
      op = OPS.ADD;
      break;

    case '-':
      op = OPS.SUBTRACT;
      break;

    case '*':
      op = OPS.MULTIPLY;
      break;

    case '/':
      op = OPS.DIVIDE;
      break;

    default:
      throw new IllegalArgumentException("Unknown Mill op");
    }
    axisIdx = 0;
  }  // end of setOp()


  //public void print()
  //{  System.out.print(axisVal);  }


  public void println()
  {  System.out.println(axisVal);  }


  public void putchar()
  // ad: used by the new "U" card
  {  
    int val = axisVal.intValue();
    if (((val >= 32) && (val < 127)) || 
         Character.isWhitespace((char)val) )
      System.out.print((char)val);
    else 
      System.out.print(' ');
  }  // end of putchar()



  public void printDouble(int numDPs)
  /* ad: print the mill value as a double its
     decimal place shifted left by numDPs  */
  {  
    double d = (new BigDecimal(axisVal, numDPs)).doubleValue();
    if(d == (long) d)
      System.out.print((long)d);  // so no nasty ".0"
    else
      System.out.print(d);
  }  // end of printDouble()


  // ---------- transfer ops --------------


  public void transferIn(BigInteger v)
  // put v into the current ingress axis (input register)
  {  transferIn(v, false);  }


  public void transferIn(BigInteger v, boolean isPrime)
  // put v into the current ingress axis (input register);
  // use the primed ingress axis if the boolean is true
  {
    if (isPrime) {
      axisVal = v;
      setIngress(2, axisVal);
    }
    else {
      axisVal = v;
      setIngress(axisIdx, axisVal);
      // When first ingress axis set, clear prime axis
      if (axisIdx == 0)
        setIngress(2, BigInteger.ZERO);

      axisIdx = (axisIdx + 1) % 2; // move to next ingress axis
      if (axisIdx == 0)  // all input registers filled
        crank();
    }
  }  // end of transferIn()


  public BigInteger transferOut()
  // get value from the egress axis (output register)
  {  return transferOut(false);  }


  public BigInteger transferOut(boolean isPrime)
  // get value from the egress axis (output register);
  // use the primed egress axis if the boolean is true
  {
    BigInteger b = egress[isPrime ? 1 : 0];
    axisVal = b;
    return b;
  }



  public void crank()
  // execute a calculation (+,-,*,/)
  {
    BigInteger result = null;
    BigInteger[] qr = null;
        // will hold the quotient and remainder of a division

    runUp = false; // Reset run up lever

    switch (op) {

    case ADD:
      result = ingress[0].add(ingress[1]);

      /* Check for passage through infinity (carry out)
       and set run up lever if that has occurred.  The
       result is then taken modulo 10^50. */
      if (result.compareTo(K10e50) >= 0) {
        runUp = true;
        result = result.subtract(K10e50);
      }
      else if ((ingress[0].signum() >= 0) && (result.signum() < 0)) {
        /* Run up gets set when the result of a
         addition changes the sign of the
         first argument.  Note that since the same
         lever is used to indicate carry and change
         of sign, it is not possible to distinguish
         overflow from sign change. */
        runUp = true;
      }
      setEgress(1, BigInteger.ZERO);
      if (isTracing) {
        System.out.println(
            "Mill:  " + ingress[0] + " + " + 
                        ingress[1] + " = " + result +
            (runUp ? " Run up" : ""));
      }
      break;


    case SUBTRACT:
      result = ingress[0].subtract(ingress[1]);

      /* Check for passage through negative infinity
       (borrow) and set run up and trim value as a
       result. */
      if (result.compareTo(Km10e50) <= 0) {
        runUp = true;
        result = result.add(K10e50).negate();
      }
      else if ((ingress[0].signum() >= 0) && (result.signum() < 0)) {
        /* Run up gets set when the result of a
         subtraction changes the sign of the
         first argument.  Note that since the same
         lever is used to indicate borrow and change
         of sign, it is not possible to distinguish
         overflow from sign change. */
        runUp = true;
      }
      if (isTracing)
        System.out.println(
            "Mill:  " + ingress[0] + " - " + 
                        ingress[1] + " = " + result +
            (runUp ? " Run up" : ""));
      setEgress(1, BigInteger.ZERO);
      break;


    case MULTIPLY:
      result = ingress[0].multiply(ingress[1]);
      BigInteger tempResult = null;
      if (isTracing)
        tempResult = result;

      /* Check for product longer than one column (50-digits) and
         set the primed egress axis to the upper part. */
      if (result.abs().compareTo(K10e50) > 0) {
        qr = result.divideAndRemainder(K10e50);
        // the array holds the quotient and remainder as BigIntegers
        setEgress(1, qr[0]);
        result = qr[1];
      }
      else
        setEgress(1, BigInteger.ZERO);

      if (isTracing)
        System.out.println(
            "Mill:  " + ingress[0] + " * " + 
                        ingress[1] + " = " + tempResult +
            (runUp ? " Run up" : ""));
      break;


    case DIVIDE:
      BigInteger dividend = ingress[0];
      if (ingress[2].signum() != 0)
        dividend = dividend.add(ingress[2].multiply(K10e50));

      if (ingress[1].signum() == 0) {
        result = BigInteger.ZERO;
        setEgress(1, result);
        runUp = true;
        break;
      }

      qr = dividend.divideAndRemainder(ingress[1]);
        // the array holds the quotient and remainder as BigIntegers
      if (qr[0].abs().compareTo(K10e50) > 0) {
        // Overflow if quotient more than 50 digits
        result = BigInteger.ZERO;
        setEgress(1, result);
        runUp = true;
        break;
      }
      setEgress(1, qr[0]);
      result = qr[1];
      if (isTracing)
        System.out.println("Mill:  " + dividend + " / " + 
                ingress[1] + " = " + qr[0] +
                ", Rem: " + qr[1] + (runUp ? " Run up" : ""));
      break;


    case NONE:
      result = axisVal;
      break;
    }

    axisVal = result;
    setEgress(0, axisVal);
    axisIdx = 0;
  }  // end of crank()



  public void shiftAxes(int count)
  // used for processing < amd > and their shift count
  {
    boolean isRight = (count < 0);
    BigInteger sf = BigInteger.ONE;
    BigInteger value;
    BigInteger[] qr = null;
        // will hold the quotient and remainder of a division


    if (isRight) {
      count = -count;
      value = egress[0];
      if (egress[1].signum() != 0)
        value = value.add(egress[1].multiply(K10e50));
    }
    else {
      value = ingress[0];
      if (ingress[2].signum() != 0)
        value = value.add(ingress[2].multiply(K10e50));
    }

    /* If we don't have a ready-made shift factor in stock,
       create one, and store it in shiftFactor[] for 
       later use.  */
    if (shiftFactor[count] == BigInteger.ONE) {
      sf = BigInteger.ONE;
      for (int j = 0; j < count; j++)
        sf = sf.multiply(BigInteger.TEN);
      shiftFactor[count] = sf;
    }

    /* Perform the shift and put the result back where
       we got it.  */
    sf = shiftFactor[count];
    if (isRight) {  // shifting right
      qr = value.divideAndRemainder(sf);
        // the array holds the quotient and remainder as BigIntegers
      if (isTracing)
        System.out.println("Mill:  " + value + " > " + 
                                       count + " = " + qr[0]);
      if (qr[0].compareTo(K10e50) != 0) {
        qr = qr[0].divideAndRemainder(K10e50);
        setEgress(1, qr[0]);
        setEgress(0, qr[1]);
      }
      else {
        setEgress(0, qr[0]);
        setEgress(1, BigInteger.ZERO);
      }
      axisVal = egress[0];
    }
    else {   // shifting left
      BigInteger pr = value.multiply(sf);
      if (isTracing)
        System.out.println("Mill:  " + value + " < " + 
                                       count + " = " + pr);
      if (pr.compareTo(K10e50) != 0) {
        qr = pr.divideAndRemainder(K10e50);
        // the array holds the quotient and remainder as BigIntegers
        setIngress(2, qr[0]);
        setIngress(0, qr[1]);
      }
      else {
        setIngress(0, pr);
        setIngress(2, BigInteger.ZERO);
      }
      axisVal = ingress[0];
    }
  }  // end of shiftAxes()


}  // end of Mill class
