
// Store.java

/* Fourmilab's Analytical Engine, by John Walker
   https://www.fourmilab.ch/babbage/contents.html

   Cut down functionality by:
   Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2018

   The store corresponds to RAM, and is made up of a 'rack'
   of 1000 'columns'. Each column is a memory cell that
   can hold a 50-digit decimal number. In earlier writings, 
   Babbage used 40-digit numbers

   In the original code, a BigInt class was used rather than
   Java's BigInteger.
*/

import java.util.*;
import java.math.*;



public class Store
{
  private final static BigInteger K10e50 = new BigInteger(
      "100000000000000000000000000000000000000000000000000");
      // a memory cell can hold up to 50 decimal digits, 
      // so this is a max value for a number

  private static final int MAX_VARS = 1000;


  private BigInteger[] rack; 
  private boolean isTracing = false;


  public Store(boolean t)
  {
     rack = new BigInteger[MAX_VARS];
     for(int i=0; i < rack.length; i++)
       rack[i] = BigInteger.ZERO;
     isTracing = t;
  }


  public void set(int idx, long v)
  {  set(idx, BigInteger.valueOf(v));  }


  public void set(int idx, BigInteger v)
  {
    if ((idx < 0) || (idx >= MAX_VARS))
      System.out.println("Store index out of range: " + idx);
    else if (v.abs().compareTo(K10e50) > 0)  // too big
      System.out.println("Value too large: " + v);
    else {
      rack[idx] = v;
      if (isTracing)
        System.out.println("Store: V" + idx + " = " + v);
    }
  }  // end of set()


  public BigInteger get(int idx)
  {
    if ((idx < 0) || (idx >= MAX_VARS)) {
      System.out.println("Store index out of range: " + idx);
      return null;
    }
    else {
      BigInteger v = rack[idx];
      if (isTracing)
        System.out.println("Store: Mill <= V" + idx + "(" + v + ")");
      return v;
    }
  }  // end of get()

}  // end of Store class
